﻿using System;
using System.Collections.Concurrent;
using TcHmiSrv;
using TcHmiSrv.Management;

namespace EventExtensionSample
{
    // The default class of the server extension must inherit from the IExtension interface.
    // The path of the server extension is available via the ExtensionPath property if the default class of the server extension inherits from the ExtensionWithPath base class instead.
    public class EventExtensionSample : IExtension
    {
        private ITcHmiSrvExtHost _host = null;

        private object _shutdownLock = new object();
        private bool _isShuttingDown = false;

        private TcHmiSrvRequestListener _requestListener = new TcHmiSrvRequestListener();
        private TcHmiSrvShutdownListener _shutdownListener = new TcHmiSrvShutdownListener();
        private TcHmiSrvAlarmProviderListener _alarmProviderListener = new TcHmiSrvAlarmProviderListener();

        private Log _logger = null;

        private ConcurrentDictionary<int, Alarm> _alarms = new ConcurrentDictionary<int, Alarm>();
        private int _alarmCounter = 0;
        private object _alarmCounterLock = new object();

        // Initializes the server extension.
        public ErrorValue Init(ITcHmiSrvExtHost host, Context context)
        {
            _logger = new Log(host);

            try
            {
                _host = host;

                // Add event handlers
                _requestListener.OnRequest += OnRequest;
                _shutdownListener.OnShutdown += OnShutdown;
                _alarmProviderListener.OnInitializeAlarmList += OnInitializeAlarmList;
                _alarmProviderListener.OnListEvents += OnListEvents;
                _alarmProviderListener.OnConfirm += OnConfirm;

                // Register listeners
                _host.register_listener(context, _requestListener);
                _host.register_listener(context, _shutdownListener);
                _host.register_listener(context, _alarmProviderListener);

                _logger.Send(context, "MESSAGE_INIT");
                return ErrorValue.HMI_SUCCESS;
            }
            catch (Exception ex)
            {
                _logger.Send(context, "ERROR_INIT", ex.Message, Severity.Severity_Error);
                return ErrorValue.HMI_E_EXTENSION_LOAD;
            }
        }

        // Called when a client requests a symbol from the extension domain.
        private ErrorValue OnRequest(object sender, TcHmiSrvRequestListener.OnRequestEventArgs e)
        {
            ErrorValue ret = ErrorValue.HMI_SUCCESS;
            Context context = e.Context;
            CommandGroup commands = e.Commands;

            try
            {
                commands.Result = ExtensionErrorValue.HMI_EXT_SUCCESS;
                string mapping = "";

                foreach (Command command in commands)
                {
                    mapping = command.Mapping;

                    try
                    {
                        // Use the mapping to check which command is requested
                        switch (mapping)
                        {
                            case "SendMessage":
                                ret = SendMessage(context, command);
                                break;

                            case "RaiseAlarm":
                                ret = RaiseAlarm(context, command);
                                break;

                            case "ClearAlarm":
                                ret = ClearAlarm(context, command);
                                break;

                            default:
                                ret = ErrorValue.HMI_E_EXTENSION;
                                break;
                        }

                        //if (ret != ErrorValue.HMI_SUCCESS)
                        // Do something on error
                    }
                    catch (Exception ex)
                    {
                        command.ExtensionResult = ExtensionErrorValue.HMI_EXT_FAIL;
                        command.ResultString = _logger.Localize(context, "ERROR_CALL_COMMAND", new string[] { mapping, ex.Message });
                    }
                }
            }
            catch
            {
                commands.Result = ExtensionErrorValue.HMI_EXT_FAIL;
            }
            finally
            {
                if (commands.Result != ExtensionErrorValue.HMI_EXT_SUCCESS)
                {
                    // Reset the read value of the commands to prevent the server from sending invalid data
                    foreach (Command command in commands)
                    {
                        command.ReadValue = null;
                    }

                    ret = ErrorValue.HMI_E_EXTENSION;
                }
            }

            return ret;
        }

        // Called when the server is shutting down. After exiting this function the extension dll will be unloaded.
        private ErrorValue OnShutdown(object sender, TcHmiSrvShutdownListener.OnShutdownEventArgs e)
        {
            // If the extension does not shutdown after 10 seconds (blocking threads) OnShutdown will be called again
            lock (_shutdownLock)
            {
                if (_isShuttingDown)
                {
                    return ErrorValue.HMI_SUCCESS;
                }

                _isShuttingDown = true;

                Context context = e.Context;

                try
                {
                    // Unregister listeners
                    _host.unregister_listener(context, _requestListener);
                    _host.unregister_listener(context, _shutdownListener);
                    _host.unregister_listener(context, _alarmProviderListener);

                    _logger.Send(context, "MESSAGE_SHUTDOWN");
                    return ErrorValue.HMI_SUCCESS;
                }
                catch (Exception ex)
                {
                    _logger.Send(context, "ERROR_SHUTDOWN", ex.Message, Severity.Severity_Error);
                    return ErrorValue.HMI_E_EXTENSION;
                }
            }
        }

        private ErrorValue OnInitializeAlarmList(object sender, TcHmiSrvAlarmProviderListener.OnInitializeAlarmListEventArgs e)
        {
            _alarms = new ConcurrentDictionary<int, Alarm>();

            return ErrorValue.HMI_SUCCESS;
        }

        private ErrorValue OnListEvents(object sender, TcHmiSrvAlarmProviderListener.OnListEventsEventArgs e)
        {
            e.Events.Type = TcHmiSrv.ValueType.ValueType_Vector;

            foreach (Alarm alarm in _alarms.Values)
            {
                e.Events.add(alarm);
            }

            return ErrorValue.HMI_SUCCESS;
        }

        private ErrorValue OnConfirm(object sender, TcHmiSrvAlarmProviderListener.OnConfirmEventArgs e)
        {
            if (!_alarms.ContainsKey(e.Alarm.Id))
            {
                return ErrorValue.HMI_E_EXTENSION;
            }

            Alarm alarm = _alarms[e.Alarm.Id];

            if (alarm.ConfirmationState == AlarmConfirmationState.AlarmConfirmationState_Wait)
            {
                alarm.ConfirmationState = AlarmConfirmationState.AlarmConfirmationState_Confirmed;
                alarm.TimeConfirmed = Timestamp.Now;

                AlarmChangeType changeType = AlarmChangeType.AlarmChangeType_Change;

                if (alarm.TimeCleared.Nanoseconds != 0)
                {
                    _alarms.TryRemove(alarm.Id, out _);
                    changeType = AlarmChangeType.AlarmChangeType_Dispose;
                }

                _host.notify_listener(e.Context, changeType, alarm);
            }

            return ErrorValue.HMI_SUCCESS;
        }

        private ErrorValue RaiseAlarm(Context context, Command command)
        {
            Alarm alarm = new Alarm(context, "SAMPLE_ALARM");
            alarm.Severity = Severity.Severity_Error;
            alarm.ConfirmationState = AlarmConfirmationState.AlarmConfirmationState_Wait;
            
            lock (_alarmCounterLock)
            {
                alarm.Id = _alarmCounter++;
            }

            _alarms.TryAdd(alarm.Id, alarm);

            _host.notify_listener(context, AlarmChangeType.AlarmChangeType_Raise, alarm);

            return ErrorValue.HMI_SUCCESS;
        }

        private ErrorValue ClearAlarm(Context context, Command command)
        {
            if (!_alarms.ContainsKey(command.WriteValue))
            {
                return ErrorValue.HMI_E_EXTENSION;
            }

            Alarm alarm = _alarms[command.WriteValue];

            if (alarm.TimeCleared.Nanoseconds == 0)
            {
                alarm.TimeCleared = Timestamp.Now;

                AlarmChangeType changeType = AlarmChangeType.AlarmChangeType_Change;

                if (alarm.TimeConfirmed.Nanoseconds != 0)
                {
                    _alarms.TryRemove(alarm.Id, out _);
                    changeType = AlarmChangeType.AlarmChangeType_Dispose;
                }

                _host.notify_listener(context, changeType, alarm);
            }

            return ErrorValue.HMI_SUCCESS;
        }

        private ErrorValue SendMessage(Context context, Command command)
        {
            Message message = new Message(context, Severity.Severity_Info, "SAMPLE_MESSAGE");

            _host.send(context, message);

            return ErrorValue.HMI_SUCCESS;
        }
    }
}
